from os import makedirs
import torch
import pickle
from itertools import product
from torch.utils.data import DataLoader
from sklearn.datasets import load_svmlight_file
from sklearn.preprocessing import normalize
import torch.utils.data as data_utils
import os
from zero_order_gradients import zo_grad_forward, zo_grad_central, zo_grad_forward_coord, zo_grad_central_coord


def svm(x, data, target, *args):
    return torch.nn.functional.relu(data.mv(x).mul(target)).mean()


def svm_grad(x, data, target):
    with torch.no_grad():
        scale = torch.nn.functional.relu(data.mv(x).mul(target))
    return data.t().mul(target.mul(torch.sign(scale))).mean(dim=1)


def main():

    dataset_name = "a9a"
    filename = "./data/" + dataset_name + ".txt"
    dataset = load_svmlight_file(filename)
    device = torch.device('cpu')
    test_size = 6561
    data_n = normalize(dataset[0].toarray(), norm='l2', axis=1)
    test_data = torch.tensor(data_n[:test_size], dtype=torch.float, device=device)
    data = torch.tensor(data_n[test_size:], dtype=torch.float, device=device)
    target_n = dataset[1]
    test_target = torch.tensor(target_n[:test_size], dtype=torch.float, device=device)
    target = torch.tensor(target_n[test_size:], dtype=torch.float, device=device)
    train_load = data_utils.TensorDataset(data, target)
    #test_load = data_utils.TensorDataset(test_data, test_target)
    dimension = data.size()[1]
    results_dir = os.environ["RESULTS"]

    method = torch.optim.SGD
    method_name = "SGD"
    gradients = [zo_grad_forward, zo_grad_central, zo_grad_forward_coord, zo_grad_central_coord, "grad"]

    reg = os.environ["REG"]
    reg = float(reg)

    bs = os.environ["BS"]
    batch_size = int(bs)
    lrs = [0.4, 0.1, 0.025, 0.00625, 0.0016, 0.0004]
    tau = os.environ["TAU"]
    tau = float(tau)

    epochs = 50
    for gradient, lr in product(gradients, lrs):
        if gradient is zo_grad_forward:
            gradient_name = "forward"
        elif gradient is zo_grad_central:
            gradient_name = "central"
        elif gradient is zo_grad_forward_coord:
            gradient_name = "forward_coord"
        elif gradient is zo_grad_central_coord:
            gradient_name = "central_coord"
        else:
            gradient_name = "grad"


        directory = f"{results_dir}/SVM/{dataset_name}/{method_name}/{gradient_name}"
        makedirs(directory, exist_ok=True)
        torch.random.manual_seed(0)
        train_dataloader = DataLoader(train_load, batch_size=batch_size, shuffle=True)
        train_losses = []
        test_losses = []
        x = torch.randn(dimension, dtype=torch.float, device=device).requires_grad_()
        optimizer = method([x], lr=lr)

        for _ in range(epochs):
            for i, (batch_data, batch_target) in enumerate(train_dataloader):
                if i % round(26000/batch_size/13) == 0:
                    with torch.no_grad():
                        full_loss = svm(x, data, target) + (reg / 2) * torch.linalg.norm(x).square()
                        test_loss = svm(x, test_data, test_target) + (reg / 2) * torch.linalg.norm(x).square()
                        train_losses.append(full_loss)
                        test_losses.append(test_loss)

                optimizer.zero_grad()
                if gradient is "grad":
                    grad = svm_grad(x, batch_data, batch_target) + x.mul(reg)
                else:
                    def closure():
                        return svm(x, batch_data, batch_target) + (reg / 2) * torch.linalg.norm(x).square()

                    grad = gradient(closure, [x], tau)[0]

                x.grad = grad
                optimizer.step()



        file_name = f"{directory}/reg_{reg}.batch_{batch_size}.lr_{round(lr, 4)}.tau_{tau}"
        if os.path.exists(file_name):
            print('ok')
        pickle.dump([train_losses, test_losses, x.detach()], open(file_name, "wb"))


if __name__ == "__main__":
    main()
